<p><a href="{{ pathto(master_doc) }}">
  <img src="{{ pathto('_images/logo-full.jpg', 1) }}" alt="Logo" style="width:80%;height:80%"/>
</a></p>

<h3>About Pint</h3>
Units in Python.
You are currently looking at the documentation of version {{ version }}.
<h3>Other Formats</h3>
<p>
  You can download the documentation in other formats as well:
</p>
<ul>
  <li><a href="https://media.readthedocs.org/pdf/pint/latest/pint.pdf">as PDF</a>
  <li><a href="https://media.readthedocs.org/htmlzip/pint/latest/pint.zip">as zipped HTML</a>
  <li><a href="https://media.readthedocs.org/epub/pint/latest/pint.epub">as ePub</a>
</ul>
<h3>Useful Links</h3>
<ul>
  <li><a href="https://pypi.python.org/pypi/Pint/">Pint @ PyPI</a></li>
  <li><a href="https://github.com/hgrecco/pint">Code in GitHub</a></li>
  <li><a href="https://github.com/hgrecco/pint/issues">Issue Tracker</a></li>
</ul>
